#!/usr/bin/env python3
"""
Secure File Transfer
Advanced secure file transfer and encryption tool.
Parameters:
    - source: Source file or directory
    - destination: Destination path
    - mode: Transfer mode (send, receive)
    - key: Encryption key file
Requires: cryptography, paramiko, pysftp packages
"""
import argparse
import logging
from datetime import datetime
import json
import os
import sys
from pathlib import Path
import hashlib
import base64
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import paramiko
import pysftp
import socket
import threading
import queue
from tqdm import tqdm

class SecureFileTransfer:
    def __init__(self, source, destination, mode='send', key_file=None):
        self.source = Path(source)
        self.destination = Path(destination)
        self.mode = mode.lower()
        self.key_file = key_file
        self.results = {}
        self.setup_logging()
        self.initialize_crypto()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def initialize_crypto(self):
        """Initialize cryptography components"""
        try:
            if self.key_file and os.path.exists(self.key_file):
                with open(self.key_file, 'rb') as f:
                    self.key = f.read()
            else:
                self.key = Fernet.generate_key()
                if self.key_file:
                    with open(self.key_file, 'wb') as f:
                        f.write(self.key)
            
            self.cipher = Fernet(self.key)
        except Exception as e:
            self.logger.error(f"Error initializing crypto: {str(e)}")
            raise

    def encrypt_file(self, file_path):
        """Encrypt file contents"""
        try:
            with open(file_path, 'rb') as f:
                data = f.read()
            encrypted_data = self.cipher.encrypt(data)
            encrypted_path = str(file_path) + '.encrypted'
            with open(encrypted_path, 'wb') as f:
                f.write(encrypted_data)
            return encrypted_path
        except Exception as e:
            self.logger.error(f"Error encrypting file: {str(e)}")
            return None

    def decrypt_file(self, file_path):
        """Decrypt file contents"""
        try:
            with open(file_path, 'rb') as f:
                encrypted_data = f.read()
            decrypted_data = self.cipher.decrypt(encrypted_data)
            decrypted_path = str(file_path).replace('.encrypted', '')
            with open(decrypted_path, 'wb') as f:
                f.write(decrypted_data)
            return decrypted_path
        except Exception as e:
            self.logger.error(f"Error decrypting file: {str(e)}")
            return None

    def calculate_checksum(self, file_path):
        """Calculate file checksum"""
        try:
            sha256_hash = hashlib.sha256()
            with open(file_path, 'rb') as f:
                for chunk in iter(lambda: f.read(4096), b''):
                    sha256_hash.update(chunk)
            return sha256_hash.hexdigest()
        except Exception as e:
            self.logger.error(f"Error calculating checksum: {str(e)}")
            return None

    def verify_transfer(self, source_path, dest_path):
        """Verify file transfer integrity"""
        source_hash = self.calculate_checksum(source_path)
        dest_hash = self.calculate_checksum(dest_path)
        return source_hash == dest_hash

    def send_local(self, source_path, dest_path):
        """Send file locally"""
        try:
            # Encrypt file
            encrypted_path = self.encrypt_file(source_path)
            if not encrypted_path:
                return False

            # Create destination directory if needed
            dest_path.parent.mkdir(parents=True, exist_ok=True)

            # Copy encrypted file
            with open(encrypted_path, 'rb') as src, \
                 open(dest_path, 'wb') as dst:
                dst.write(src.read())

            # Clean up encrypted file
            os.remove(encrypted_path)

            return self.verify_transfer(source_path, dest_path)
        except Exception as e:
            self.logger.error(f"Error in local transfer: {str(e)}")
            return False

    def send_sftp(self, source_path, hostname, username, password=None, key_path=None):
        """Send file via SFTP"""
        try:
            # Encrypt file
            encrypted_path = self.encrypt_file(source_path)
            if not encrypted_path:
                return False

            # Setup SFTP connection
            cnopts = pysftp.CnOpts()
            cnopts.hostkeys = None  # For testing only
            
            if key_path:
                with pysftp.Connection(
                    hostname,
                    username=username,
                    private_key=key_path,
                    cnopts=cnopts
                ) as sftp:
                    sftp.put(
                        encrypted_path,
                        str(self.destination)
                    )
            else:
                with pysftp.Connection(
                    hostname,
                    username=username,
                    password=password,
                    cnopts=cnopts
                ) as sftp:
                    sftp.put(
                        encrypted_path,
                        str(self.destination)
                    )

            # Clean up encrypted file
            os.remove(encrypted_path)

            return True
        except Exception as e:
            self.logger.error(f"Error in SFTP transfer: {str(e)}")
            return False

    def receive_local(self, source_path, dest_path):
        """Receive file locally"""
        try:
            # Create destination directory if needed
            dest_path.parent.mkdir(parents=True, exist_ok=True)

            # Decrypt file
            decrypted_path = self.decrypt_file(source_path)
            if not decrypted_path:
                return False

            # Copy decrypted file
            with open(decrypted_path, 'rb') as src, \
                 open(dest_path, 'wb') as dst:
                dst.write(src.read())

            # Clean up decrypted file
            os.remove(decrypted_path)

            return self.verify_transfer(source_path, dest_path)
        except Exception as e:
            self.logger.error(f"Error in local receive: {str(e)}")
            return False

    def receive_sftp(self, source_path, hostname, username, password=None, key_path=None):
        """Receive file via SFTP"""
        try:
            # Setup SFTP connection
            cnopts = pysftp.CnOpts()
            cnopts.hostkeys = None  # For testing only
            
            if key_path:
                with pysftp.Connection(
                    hostname,
                    username=username,
                    private_key=key_path,
                    cnopts=cnopts
                ) as sftp:
                    sftp.get(
                        str(source_path),
                        str(self.destination)
                    )
            else:
                with pysftp.Connection(
                    hostname,
                    username=username,
                    password=password,
                    cnopts=cnopts
                ) as sftp:
                    sftp.get(
                        str(source_path),
                        str(self.destination)
                    )

            # Decrypt received file
            decrypted_path = self.decrypt_file(self.destination)
            if not decrypted_path:
                return False

            return True
        except Exception as e:
            self.logger.error(f"Error in SFTP receive: {str(e)}")
            return False

    def process_directory(self, source_dir, callback):
        """Process directory recursively"""
        try:
            total_files = sum(1 for _ in source_dir.rglob('*') if _.is_file())
            processed_files = 0
            failed_files = 0

            with tqdm(total=total_files, desc="Processing files") as pbar:
                for item in source_dir.rglob('*'):
                    if item.is_file():
                        relative_path = item.relative_to(source_dir)
                        dest_path = self.destination / relative_path
                        
                        if callback(item, dest_path):
                            processed_files += 1
                        else:
                            failed_files += 1
                        
                        pbar.update(1)

            return processed_files, failed_files
        except Exception as e:
            self.logger.error(f"Error processing directory: {str(e)}")
            return 0, 0

    def generate_report(self):
        """Generate transfer report"""
        try:
            report = {
                'timestamp': datetime.now().isoformat(),
                'source': str(self.source),
                'destination': str(self.destination),
                'mode': self.mode,
                'results': self.results
            }

            # Add summary section
            summary = {
                'total_files': self.results.get('total_files', 0),
                'successful_transfers': self.results.get('successful', 0),
                'failed_transfers': self.results.get('failed', 0),
                'total_bytes': self.results.get('total_bytes', 0)
            }

            report['summary'] = summary
            return report
        except Exception as e:
            self.logger.error(f"Error generating report: {str(e)}")
            return None

    def save_report(self, report, output_file):
        """Save transfer report to file"""
        try:
            with open(output_file, 'w') as f:
                json.dump(report, f, indent=4)
            self.logger.info(f"Report saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving report: {str(e)}")

    def print_summary(self, report):
        """Print transfer summary"""
        self.logger.info("\nFile Transfer Summary:")
        self.logger.info(f"Source: {report['source']}")
        self.logger.info(f"Destination: {report['destination']}")
        self.logger.info(f"Mode: {report['mode']}")
        
        if 'summary' in report:
            summary = report['summary']
            self.logger.info("\nTransfer Statistics:")
            self.logger.info(f"Total Files: {summary['total_files']}")
            self.logger.info(
                f"Successful Transfers: {summary['successful_transfers']}"
            )
            self.logger.info(
                f"Failed Transfers: {summary['failed_transfers']}"
            )
            self.logger.info(
                f"Total Bytes Transferred: {summary['total_bytes']:,}"
            )

    def start_transfer(self, output_file=None):
        """Start file transfer process"""
        self.logger.info("Starting secure file transfer")
        self.logger.info(f"Mode: {self.mode}")
        self.logger.info(f"Source: {self.source}")
        self.logger.info(f"Destination: {self.destination}")
        
        try:
            # Initialize results
            self.results = {
                'total_files': 0,
                'successful': 0,
                'failed': 0,
                'total_bytes': 0
            }

            # Process based on source type
            if self.source.is_file():
                if self.mode == 'send':
                    success = self.send_local(
                        self.source,
                        self.destination
                    )
                else:
                    success = self.receive_local(
                        self.source,
                        self.destination
                    )
                
                self.results['total_files'] = 1
                if success:
                    self.results['successful'] = 1
                    self.results['total_bytes'] = self.source.stat().st_size
                else:
                    self.results['failed'] = 1
                    
            elif self.source.is_dir():
                if self.mode == 'send':
                    processed, failed = self.process_directory(
                        self.source,
                        lambda src, dst: self.send_local(src, dst)
                    )
                else:
                    processed, failed = self.process_directory(
                        self.source,
                        lambda src, dst: self.receive_local(src, dst)
                    )
                
                self.results.update({
                    'total_files': processed + failed,
                    'successful': processed,
                    'failed': failed,
                    'total_bytes': sum(
                        f.stat().st_size for f in self.source.rglob('*')
                        if f.is_file()
                    )
                })

            # Generate and save report
            report = self.generate_report()
            if report and output_file:
                self.save_report(report, output_file)
            if report:
                self.print_summary(report)
                
        except KeyboardInterrupt:
            self.logger.info("Transfer interrupted by user")
        except Exception as e:
            self.logger.error(f"Error during transfer: {str(e)}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Secure File Transfer"
    )
    parser.add_argument(
        "source",
        help="Source file or directory"
    )
    parser.add_argument(
        "destination",
        help="Destination path"
    )
    parser.add_argument(
        "--mode",
        choices=['send', 'receive'],
        default='send',
        help="Transfer mode"
    )
    parser.add_argument(
        "--key",
        help="Encryption key file"
    )
    parser.add_argument(
        "--output",
        help="Output file for transfer report"
    )
    args = parser.parse_args()

    transfer = SecureFileTransfer(
        args.source,
        args.destination,
        args.mode,
        args.key
    )
    transfer.start_transfer(args.output)
